/* Copyright (c) 2008, Nathan Sweet
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
 * conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided with the distribution.
 * - Neither the name of Esoteric Software nor the names of its contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package dorkbox.network.kryo;

import com.esotericsoftware.kryo.Kryo;
import com.esotericsoftware.kryo.io.Input;
import com.esotericsoftware.kryo.io.Output;
import com.esotericsoftware.kryo.serializers.MapSerializer;
import org.junit.Assert;

import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Random;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;

/** @author Nathan Sweet <misc@n4te.com> */
public class MapSerializerTest extends KryoTestCase {
    {
        supportsCopy = true;
    }

    @SuppressWarnings({"rawtypes","unchecked"})
    public void testMaps () {
        kryo.register(HashMap.class);
        kryo.register(LinkedHashMap.class);
        HashMap map = new HashMap();
        map.put("123", "456");
        map.put("789", "abc");
        roundTrip(18, 21, map);
        roundTrip(2, 5, new LinkedHashMap());
        roundTrip(18, 21, new LinkedHashMap(map));

        MapSerializer serializer = new MapSerializer();
        kryo.register(HashMap.class, serializer);
        kryo.register(LinkedHashMap.class, serializer);
        serializer.setKeyClass(String.class, kryo.getSerializer(String.class));
        serializer.setKeysCanBeNull(false);
        serializer.setValueClass(String.class, kryo.getSerializer(String.class));
        roundTrip(14, 17, map);
        serializer.setValuesCanBeNull(false);
        roundTrip(14, 17, map);
    }

    public void testEmptyHashMap () {
        execute(new HashMap<Object, Object>(), 0);
    }

    public void testNotEmptyHashMap () {
        execute(new HashMap<Object, Object>(), 1000);
    }

    public void testEmptyConcurrentHashMap () {
        execute(new ConcurrentHashMap<Object, Object>(), 0);
    }

    public void testNotEmptyConcurrentHashMap () {
        execute(new ConcurrentHashMap<Object, Object>(), 1000);
    }

    public void testGenerics () {
        kryo.register(HasGenerics.class);
        kryo.register(Integer[].class);
        kryo.register(HashMap.class);

        HasGenerics test = new HasGenerics();
        test.map.put("moo", new Integer[] {1, 2});

        output = new Output(4096);
        kryo.writeClassAndObject(output, test);
        output.flush();

        input = new Input(output.toBytes());
        HasGenerics test2 = (HasGenerics)kryo.readClassAndObject(input);
        assertEquals(test.map.get("moo"), test2.map.get("moo"));
    }

    private void execute (Map<Object, Object> map, int inserts) {
        Random random = new Random();
        for (int i = 0; i < inserts; i++) {
            map.put(random.nextLong(), random.nextBoolean());
        }

        Kryo kryo = new Kryo();
        kryo.register(HashMap.class, new MapSerializer());
        kryo.register(ConcurrentHashMap.class, new MapSerializer());

        Output output = new Output(2048, -1);
        kryo.writeClassAndObject(output, map);
        output.close();

        Input input = new Input(output.toBytes());
        Object deserialized = kryo.readClassAndObject(input);
        input.close();

        Assert.assertEquals(map, deserialized);
    }

    @SuppressWarnings({"unchecked","rawtypes"})
    public void testTreeMap () {
        kryo.register(TreeMap.class);
        TreeMap map = new TreeMap();
        map.put("123", "456");
        map.put("789", "abc");
        roundTrip(19, 22, map);

        kryo.register(KeyThatIsntComparable.class);
        kryo.register(KeyComparator.class);
        map = new TreeMap(new KeyComparator());
        KeyThatIsntComparable key = new KeyThatIsntComparable();
        key.value = "123";
        map.put(key, "456");
        roundTrip(11, 14, map);
    }

    @SuppressWarnings({"unchecked","rawtypes"})
    static public class HasGenerics {
        public HashMap<String, Integer[]> map = new HashMap();
        public HashMap<String, ?> map2 = new HashMap();
    }

    static public class KeyComparator implements Comparator<KeyThatIsntComparable> {
        @Override
        public int compare (KeyThatIsntComparable o1, KeyThatIsntComparable o2) {
            return o1.value.compareTo(o2.value);
        }
    }

    static public class KeyThatIsntComparable {
        public String value;
    }
}
